#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <linux/joystick.h>

#include <time.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "joystick.h"
#include "master.h"

#include "joystick_if.h"
#include "none_config.h"

extern int bpm, beat;
extern int tick;

int joystick_calc_note(joystick * js);

module_class joystick_mif = {
  SEQUENCER,
  0,
  (char *) "joystick",
  (char *) "Joystick",
  NULL,
  (void *) joystick_new,
  (void *) joystick_clone,
  NULL,
  NULL,
  (void *) none_config_cb,
  (void *) joystick_if_new,
  (void *) joystick_if_close_cb
};

joystick *
joystick_new()
{
  joystick *js;

  js = (joystick *) malloc(sizeof(joystick));
  js->module.class = &joystick_mif;
  js->module.on = 0;
  js->module.nr_inputs = 0;
  js->module.nr_outputs = 3;
  js->module.outputs[0] = malloc(sizeof(channel));
  js->module.outputs[0]->parms.ch_type = CH_TYPE_SEQUENCE;
  js->module.outputs[0]->data = (seq_channel *) (&(js->seq_ch[0]));
  js->module.outputs[1] = malloc(sizeof(channel));
  js->module.outputs[1]->parms.ch_type = CH_TYPE_SEQUENCE;
  js->module.outputs[1]->data = (seq_channel *) (&(js->seq_ch[1]));
  js->module.outputs[2] = malloc(sizeof(channel));
  js->module.outputs[2]->parms.ch_type = CH_TYPE_SEQUENCE;
  js->module.outputs[2]->data = (seq_channel *) (&(js->seq_ch[2]));

  aube_module_clear_outputs((module *) js);
  sprintf(js->module.u_label, "joystick");
  sprintf(js->module.outputs[0]->u_label, "0");
  js->module.outputs[0]->module = (module *) js;
  sprintf(js->module.outputs[1]->u_label, "1");
  js->module.outputs[1]->module = (module *) js;
  sprintf(js->module.outputs[2]->u_label, "2");
  js->module.outputs[2]->module = (module *) js;

  js->module.do_tick = (void *) joystick_calc_note;

  aube_add_module((module *) js);

  if((js->fd = open("/dev/js0", O_RDONLY)) < 0) {
    snprintf(js->name, "No joystick found", JOYSTICK_NAME_LENGTH);
    perror("jstest");
  } else {

    ioctl(js->fd, JSIOCGAXES, &js->axes);
    ioctl(js->fd, JSIOCGBUTTONS, &js->buttons);
    ioctl(js->fd, JSIOCGNAME(JOYSTICK_NAME_LENGTH), js->name);

    printf("Joystick (%s) has %d axes and %d buttons.\n",
           js->name, js->axes, js->buttons);
  }

  js->vol = 8;
  js->pan = 16;

  return js;
}

joystick *
joystick_clone(joystick * ojs)
{
  joystick *js;

  js = (joystick *) malloc(sizeof(joystick));
  memcpy(js, ojs, sizeof(joystick));

  js->module.nr_inputs = 0;
  js->module.nr_outputs = 1;
  js->module.outputs[0] = malloc(sizeof(channel));
  js->module.outputs[0]->parms.ch_type = CH_TYPE_SEQUENCE;
  js->module.outputs[0]->data = (bit16 *) (&(js->seq_ch));

  aube_module_clear_outputs((module *) js);
  sprintf(js->module.u_label, "joystick");
  sprintf(js->module.outputs[0]->u_label, "seq");
  js->module.outputs[0]->module = (module *) js;

  js->module.do_tick = (void *) joystick_calc_note;

  aube_add_module((module *) js);

  return js;
}

int 
joystick_calc_note(joystick * js)
{
  unsigned int i;
  int lpan, rpan;
  struct js_event ev;
  struct timeval tv;
  fd_set fds;
  int sch;
  int val;

  for(sch=0; sch<3; sch++) {
    js->seq_ch[sch].trigger = 0;      /* set to 1 when fired */
    js->seq_ch[sch].pan = js->pan;
    js->seq_ch[sch].vol = js->vol;;
    js->seq_ch[sch].accent = 0;
  }

  FD_ZERO(&fds);
  FD_SET(js->fd, &fds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  while(select(js->fd+1, &fds, NULL, NULL, &tv)) {
    if(read(js->fd, &ev, sizeof(ev)) != sizeof(ev)) {
      perror("error reading joystick");
    } else if (js->module.on) {
      switch(ev.type & ~JS_EVENT_INIT) {
      case JS_EVENT_BUTTON:
        if(0<= ev.number && ev.number < 3)
          js->seq_ch[ev.number].trigger = ev.value;
        break;
      case JS_EVENT_AXIS:
        if(ev.number == 4) { /* Y-axis on Cyberman 2 */
          sch = 0;
          val = -ev.value;
        } else if (ev.number == 5) {
          sch = 1;
          val = ev.value;
        } else if (ev.number == 3) {
          sch = 2;
          val = ev.value;
        }

        js->seq_ch[sch].pitch = (double)(((val + 16384) >> 10) + 100);
        if(js->seq_ch[sch].pitch < 0.0) {
          js->seq_ch[sch].pitch = 0.0;
        }

	/*printf("value: %d\tpitch: %f\n", ev.value, js->seq_ch[sch].pitch);*/
        break;
      }

      for(sch=0; sch<3; sch++) {
        js->seq_ch[sch].vol = js->vol;
        js->seq_ch[sch].pan = 16;
        js->seq_ch[sch].accent = 0;
      }
    }
  }

  return PROC_SUCCESS;
}
